/* JOSX - RCX Abstract Port
Copyright (C) 2005 Georg Klima, Brian Bagnall, Lawrie Griffiths

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

Contact:
Georg Klima, gklima - at - cosy.sbg.ac.at
University of Salzburg, Austria
*/
package josx.rcxcomm;

import java.io.DataOutput;
import java.io.IOException;
import java.io.InputStream;



/** RCXAbstractPort provides an interface similar to java.net.Socket
 * Adapted from original code created by the LEGO3 Team at DTU-IAU
 * RCXAbstractPort implements input and output stream handling and input
 * buffering. It uses a packet handler for sending and receivng packets.
 * This version is abstract because it has no packet handler defined.
 * Specific versions of RCXAbstractPort override the constructor and
 * set up the packet handler to use a specific protocol stack. 
 * @author Brian Bagnall
 * @author Lawrie Griffiths
 * @author Georg Klima
 */
public abstract class RCXAbstractPort {
	
	private boolean portOpen = true;
	private Listener listener;
	private int timeOut = 0;
	
	private RCXInputStream rcxin;
	private RCXDataOutput rcxout;
	protected PacketHandler packetHandler;	
	
	/**
	 * Constructor for the RCXAbstractPort.
	 * Opens the port, and sets the protocol packet handler.
	 * @param handler the packet handler
	 */
	public RCXAbstractPort(PacketHandler handler) throws IOException {
		packetHandler = handler;
		rcxin = new RCXInputStream(this);
		rcxout = new RCXDataOutput(packetHandler);
		listener = new Listener();
		listener.setDaemon(true);
		listener.start();
	}
	
	/**
	 * Constructor for a named port (eg com1 or usb).
	 * The port name is ignored on the RCX.
	 * @param port the port name, eg com1 or usb
	 * @param handler, the packet handler
	 */
	public RCXAbstractPort(String port, PacketHandler handler) throws IOException {
		this(handler);
	}
	
	/**
	 * Switch listening on and off on the PC, for a serial tower.
	 * @param listen true to listen, else false 
	 **/
	public void setListen(boolean listen) {
		packetHandler.setListen(listen);
	}
	
	/** Returns an input stream for this RCXPort.
	 * @return an input stream for reading bytes from this RCXPort.
	 */
	public InputStream getInputStream() {
		packetHandler.setListen(true);
		return (InputStream) rcxin;
	}
	
	/** Returns an output stream for this RCXPort.
	 * @return an output stream for writing bytes to this RCXPort.
	 */
	public DataOutput getDataOutput() {
		return (DataOutput) rcxout;
	}
	
	/**
	 * Resets sequence numbers for this port 
	 */
	public void reset() {
		packetHandler.reset();
	}
	
	/** Closes this RCXPort, stopping the Listener thread.
	 */
	public void close() {
		portOpen = false;
		packetHandler.close();
	}
	
	/** Getter for property timeOut.
	 * @return Value of property timeOut.
	 */
	public int getTimeOut() {
		return timeOut;
	}
	
	/** Setter for property timeOut.
	 * @param timeOut New value of property timeOut.
	 */
	public void setTimeOut(int timeOut) {
		this.timeOut = timeOut;
	}
	
	private byte [] inPacket = new byte[260];
	
	
	//TODO See if we need some giotto in here!
	/** Listener class runs a thread that reads and buffers bytes.
	 * Allows a maximum of two bytes in a packet
	 */
	private class Listener extends Thread {
		public void run() {
			while (portOpen) {
				if (packetHandler.isPacketAvailable()) {
					int r = packetHandler.receivePacket(inPacket);
					for(int i=0;i<r;i++) rcxin.add(inPacket[i]);
				}
				try {
					Thread.sleep(10);
				} catch (InterruptedException iE) { }
			}
		}
	}
	
	/**
	 * Hidden inner class extending InputStream. 
	 */
	private class RCXInputStream extends InputStream {
		
		/** The default buffer size for the InputStream
		 */
		public static final int bufferSize = 32;
		private byte[] buffer = new byte[bufferSize];
		private int current = 0, last = 0;
		private RCXAbstractPort dataPort;
		
		/** Creates new RCXInputStream
		 * @param port The RCXAbstractPort which should deliver data for to this InputStream
		 */
		public RCXInputStream(RCXAbstractPort port) {
			dataPort = port;
		}
		
		/** Checks if there is any data avaliable on the InputStream
		 * @throws IOException is never thrown
		 * @return The number of bytes avaliable on the InputStream
		 */
		public int available() throws IOException {
			if (last < current)
				return bufferSize-(current-last);
			else
				return last-current;
		}
		
		/** Read a single byte from the InputStream. Returns value as
		 * an int value between 0 and 255.
		 * @throws IOException is thrown when the read is timed out
		 * @return A data byte from the stream
		 */
		public synchronized int read() throws IOException {
			long time1 = System.currentTimeMillis();
			long timeOut = dataPort.getTimeOut();
			while (available() == 0) {
				if (timeOut != 0 && (System.currentTimeMillis()-time1 > timeOut)) {
					throw new IOException("The read timed out");
				}
				try {
					Thread.sleep(10);
				} catch (InterruptedException iE) { }
			}
			
			synchronized (buffer) {
				int b = buffer[current++];
				if (current == bufferSize)
					current = 0;
				
				if(b < 0) b = b + 256;
				return b;
			}
		}
		
		/** Add a data byte to the stream
		 * This method should only be called by the RCXPort that
		 * created the RCXInputStream
		 * @param b The data byte
		 */
		void add(byte b) {
			synchronized (buffer) {
				buffer[last++] = b;
				if (last == bufferSize)
					last = 0;
			}
		}
	}
	
	/** 
	 * Hidden inner class implementing java.io.DataOutput. 
	 */
	private class RCXDataOutput implements DataOutput {
		
		private PacketHandler packetHandler;
		
		/** Creates new RCXOutputStream
		 * @param handler the packet handler used to send data
		 */
		public RCXDataOutput(PacketHandler handler) {
			packetHandler = handler;
		}
		
		public void write(int b) throws IOException {
			if (!packetHandler.sendPacket(new byte[]{(byte) b},1)) 
				throw new IOException("Failed to Receive Reply");
		}
		
		public void write(byte[] b) throws IOException {
			if (!packetHandler.sendPacket(b,b.length)) 
				throw new IOException("Failed to Receive Reply");
		}
		
		public void write(byte[] b, int off, int len) throws IOException {
			byte [] t = new byte[len];
			System.arraycopy(b,off,t,0,len);
			if (!packetHandler.sendPacket(t,t.length)) 
				throw new IOException("Failed to Receive Reply");
		}
		
		public void writeBoolean(boolean arg0) throws IOException {
			throw new IOException("Not yet implemented.");
		}
		
		public void writeByte(int v) throws IOException {
			if (!packetHandler.sendPacket(new byte[]{(byte) v},1)) 
				throw new IOException("Failed to Receive Reply");
		}
		
		public void writeShort(int v) throws IOException {
			write(new byte [] {(byte)(0xff & (v >> 8)), (byte)(0xff & v)});			
		}
		
		public void writeChar(int v) throws IOException {
			writeShort(v);			
		}
		
		public void writeInt(int v) throws IOException {
			write(new byte [] {	(byte)(0xff & (v >> 24)),
					(byte)(0xff & (v >> 16)),
					(byte)(0xff & (v >> 8)),
					(byte)(0xff & v)});				
		}
		
		public void writeLong(long v) throws IOException {
			write(new byte [] {	(byte)(0xff & (v >> 56)),
					(byte)(0xff & (v >> 48)),
					(byte)(0xff & (v >> 40)),
					(byte)(0xff & (v >> 32)),
					(byte)(0xff & (v >> 24)),
					(byte)(0xff & (v >> 16)),
					(byte)(0xff & (v >>  8)),
					(byte)(0xff & v)});			
		}
		
		public void writeFloat(float arg0) throws IOException {
			throw new IOException("Not yet implemented.");
		}
		
		public void writeDouble(double arg0) throws IOException {
			throw new IOException("Not yet implemented.");
		}
		
		public void writeBytes(String arg0) throws IOException {
			throw new IOException("Not yet implemented.");
		}
		
		public void writeChars(String arg0) throws IOException {
			throw new IOException("Not yet implemented.");
		}
		
		public void writeUTF(String arg0) throws IOException {
			throw new IOException("Not yet implemented.");
		}
	}
}
